/*
 *  video_thread.c
 */

/* Standard Linux headers */
#include <stdio.h>                               // always include stdio.h
#include <stdlib.h>                              // always include stdlib.h
#include <string.h>                              // defines memset and memcpy methods

/* Codec Engine headers */
#include <xdc/std.h>                             // xdc definitions. Must come 1st
#include <ti/sdo/ce/Engine.h>                    // Required for any CE application

/* DMAI headers */
#include <ti/sdo/dmai/Dmai.h>
#include <ti/sdo/dmai/Capture.h>
#include <ti/sdo/dmai/Display.h>
#include <ti/sdo/dmai/Buffer.h>
#include <ti/sdo/dmai/BufferGfx.h>
#include <ti/sdo/dmai/ce/Vdec2.h>
#include <ti/sdo/dmai/ce/Venc1.h>

/* Application header files */
#include "debug.h"                               // DBG and ERR macros
#include "video_thread.h"                        // Video thread definitions

/* Video encoder and decoder used */
#define VIDEO_ENCODER "h264enc"
#define VIDEO_DECODER "h264dec"
//#define VIDEO_ENCODER "mpeg4enc"
//#define VIDEO_DECODER "mpeg4dec"

/* Align buffers to this cache line size (in bytes)*/
#define BUFSIZEALIGN        128

/* Macro for clearing structures */
#define CLEAR(x) memset (&(x), 0, sizeof (x))

/* Numbers of video buffers needed */
#define NUM_CAPTURE_BUFS 3

/* Create a joint buffer table to be shared between decoder and display */
//#define NUM_DISPLAY_BUFS 2
//#define NUM_DECODER_BUFS 3
#define NUM_DECODER_DISPLAY_BUFS 4


/**************************************************************************
 * video_thread_fxn                                                       *
 **************************************************************************/
/*  input parameters:                                                    */
/*      void *envByRef       --  pointer to video_thread_env structure   */
/*                               as defined in video_thread.h            */
/*                                                                       */
/*          envByRef.quit    -- when quit != 0, thread will exit         */
/*          env->engineName -- string value specifying name of           */
/*                              engine configuration for Engine_open     */
/*                                                                       */
/*  return value:                                                        */
/*      void *     --  VIDEO_THREAD_SUCCESS or VIDEO_THREAD_FAILURE as   */
/*                     defined in video_thread.h                         */
/*************************************************************************/
void *video_thread_fxn( void *envByRef )
{
/*  Thread parameters and return value */
    video_thread_env *env    = envByRef;                     // see above
    void             *status = VIDEO_THREAD_SUCCESS;         // see above

/*  The levels of initialization for initMask */
#define OSDSETUPCOMPLETE         0x1
#define DISPLAYDEVICEINITIALIZED 0x2
#define CAPTUREDEVICEINITIALIZED 0x4
#define VIDEOENCODERCREATED      0x8
#define VIDEODECODERCREATED      0x10
#define ENCODEDBUFFERALLOCATED   0x20
#define ENGINEOPENED             0x40
    unsigned int initMask = 0x0;            // Used to only cleanup items that were init'ed

/*  Capture variables */
    Capture_Attrs     cAttrs = Capture_Attrs_OMAP3530_DEFAULT;
    Capture_Handle    hCapture = NULL;
    Buffer_Handle     cBuf;

/*  Display variables */
    Display_Attrs     dAttrs = Display_Attrs_O3530_VID_DEFAULT;
    Display_Handle    hDisplay = NULL;
    Buffer_Handle     dBuf;

/*  Intermediate buffer for encoded video */
    Buffer_Handle      procBuf        = NULL;
    Buffer_Handle      encBuf         = NULL;                   // pointer to encoded buffer
    Buffer_Attrs       encBufAttrs    = Buffer_Attrs_DEFAULT;
    int                encBufSize     = 0;                      // size of encoded buffer

/*  Codec engine variables */
    Engine_Handle engineHandle          = NULL;               // handle to the engine
    Venc1_Handle	      encoderHandle = NULL;                 // handle to video encoder
    Vdec2_Handle          decoderHandle = NULL;                 // handle to video decoder
    VIDENC1_Params        eParams       = Venc1_Params_DEFAULT;
    VIDDEC2_Params        dParams       = Vdec2_Params_DEFAULT;
    VIDENC1_DynamicParams eDynParams    = Venc1_DynamicParams_DEFAULT;
    VIDDEC2_DynamicParams dDynParams    = Vdec2_DynamicParams_DEFAULT;

    Int			ret		= Dmai_EOK;

    BufferGfx_Attrs  gfxAttrs = BufferGfx_Attrs_DEFAULT;
    BufTab_Handle    hBufTabCapture  = NULL, hBufTabDecoderDisplay = NULL;
    VideoStd_Type    videoStd;
    Int32            bufSize;

    BufferGfx_Dimensions   dim;

/* Thread Create Phase -- secure and initialize resources */

/* Setup Capture */

    /* Detect type of video (i.e. video standard) connected to the EVM's video input */
    if ( Capture_detectVideoStd( NULL, &videoStd, &cAttrs ) < 0 ) 
    {
        printf( "Failed to detect input video standard, input connected?\n" );
        goto cleanup;
    }

    /* Calculate buffer size needed given a color space */
    bufSize = BufferGfx_calcSize( videoStd, ColorSpace_UYVY );
    bufSize = ( bufSize + 4096) & ( ~0xFFF );

    //DBG( "Calculated Capture bufsize %d\n", bufSize );

    if ( bufSize < 0 ) 
    {
        printf( "Failed to calculate size for video driver buffers\n" );
        goto cleanup;
    }

    /* Calculate the dimensions of the video standard */
    if ( BufferGfx_calcDimensions( videoStd,
                                   ColorSpace_UYVY, 
                                   &gfxAttrs.dim ) 
         < 0 ) 
    {
        printf( "Failed to calculate dimensions for video driver buffers\n" );
        goto cleanup;
    }

    gfxAttrs.colorSpace = ColorSpace_UYVY;

    /* Create a table of buffers to use with the capture driver */
    hBufTabCapture = BufTab_create( NUM_CAPTURE_BUFS, 
                                    bufSize,
                                    BufferGfx_getBufferAttrs( &gfxAttrs ) );

    if ( hBufTabCapture == NULL ) 
    {
        printf( "Failed to allocate contiguous buffers\n" );
        goto cleanup;
    }

    /* Create the capture driver instance */
    cAttrs.numBufs = NUM_CAPTURE_BUFS;
    hCapture = Capture_create( hBufTabCapture, &cAttrs );

    if( hCapture == NULL ) 
    {
        ERR( "Failed Capture_create in video_thread_function\n" );
        status = VIDEO_THREAD_FAILURE;
        goto cleanup;
    }

    /* Record that capture device was opened in initialization bitmask */
    initMask |= CAPTUREDEVICEINITIALIZED;

    /* Open the codec engine                                           *
     * Note: Codec engine should be opened in each thread that uses it */
    engineHandle = Engine_open( env->engineName, NULL, NULL );

    if( engineHandle == NULL )
    {
        ERR( "Engine_setup failed in video_thread_fxn\n" );
        status = VIDEO_THREAD_FAILURE;
        goto cleanup;
    }
 
    initMask |= ENGINEOPENED;

    /* Allocate and initialize video encoder on the engine */
    /*     uses engineHandle global variable assumed set before entry */

    eParams.inputChromaFormat = XDM_YUV_422ILE;
    eParams.maxBitRate        = 4000000;
    eParams.maxWidth          = gfxAttrs.dim.width;
    eParams.maxHeight         = gfxAttrs.dim.height;

    eDynParams.targetBitRate  = 4000000;
    eDynParams.inputWidth     = gfxAttrs.dim.width;
    eDynParams.inputHeight    = gfxAttrs.dim.height;

    encoderHandle = Venc1_create( engineHandle, 
                                 VIDEO_ENCODER, 
                                 &eParams, 
                                 &eDynParams);

    if( encoderHandle == NULL )
    {
        ERR( "Video_encoder create failed in video_thread_fxn\n" );
        status = VIDEO_THREAD_FAILURE;
        goto cleanup;
    }

    initMask |= VIDEOENCODERCREATED;

    /* Allocate and initialize video decoder on the engine */

//    The important one here is forceChromaFormat, the others are equal to default
//    forceChromaFormat defaults to 4:2:0 instead of 4:2:2 and buffers aren't 
//    big enough.
        dParams.forceChromaFormat    = XDM_YUV_422ILE;
        dParams.maxWidth             = VideoStd_D1_WIDTH;
        dParams.maxHeight            = VideoStd_D1_PAL_HEIGHT;

    decoderHandle = Vdec2_create( engineHandle, 
                                 VIDEO_DECODER, 
                                 &dParams, 
                                 &dDynParams );

    if( decoderHandle == NULL )
    {
        ERR( "Video_decoder create failed in video_thread_fxn\n" );
        status = VIDEO_THREAD_FAILURE;
        goto cleanup;
    }

    // DMAI will assume that the output buffers provided to the decoder are
    //     at least the size requested by the decoder, so for consistency
    //     it is good practice to always use a decoder getOutBufSize query
    //     to determine the size of the buffer table buffers
    bufSize = Vdec2_getOutBufSize( decoderHandle );

    //DBG( "Calculated Decoder/Display bufsize %d\n", bufSize );

    /* Create a table of buffers to use with the display driver */
    hBufTabDecoderDisplay = BufTab_create( NUM_DECODER_DISPLAY_BUFS, 
                                           bufSize,
                                           BufferGfx_getBufferAttrs( &gfxAttrs ) );

    if ( hBufTabDecoderDisplay == NULL ) 
    {
        printf( "Failed to allocate contiguous buffers for hBufTabDecoderDisplay\n" );
        goto cleanup;
    }

    Vdec2_setBufTab( decoderHandle, hBufTabDecoderDisplay );

    initMask |= VIDEODECODERCREATED;


    /* Set buffer size for intermediate buffer (for encoded data)          */
    /* as the size of a full PAL frame                                     */
    encBufSize = Vdec2_getInBufSize( decoderHandle );

    /* Allocate intermediate buffer */
    /* Must use contiguous buffers if passed to DSP! */
    /* (Driver buffers are automatically allocated contiguous */

    encBufAttrs.memParams.align = BUFSIZEALIGN; 
    encBufAttrs.memParams.type = Memory_CONTIGPOOL;

    encBuf = Buffer_create( Dmai_roundUp( encBufSize, BUFSIZEALIGN ),
                            &encBufAttrs );


    if ( encBuf == NULL ) 
    {
        ERR( "Failed to allocate video buffer in video_thread_fxn.\n" );
        status = VIDEO_THREAD_FAILURE;
        goto cleanup;
    }

    initMask |= ENCODEDBUFFERALLOCATED;

    DBG( "Allocated intermediate video buffer of size %d\n", encBufSize );

/* Setup Display */

    /* Set output standard to match input */
    dAttrs.videoStd = Capture_getVideoStd( hCapture );
    dAttrs.numBufs  = NUM_DECODER_DISPLAY_BUFS;
    dAttrs.rotation = 90;


    /* Create the display driver instance */
    hDisplay = Display_create( hBufTabDecoderDisplay, &dAttrs);
    if( hDisplay == NULL )
    {
        ERR( "Failed to open handle to Display driver in video_thread_fxn\n" );
        status = VIDEO_THREAD_FAILURE;
        goto cleanup;
    }

    /* Record that display device was opened in initialization bitmask */
    initMask |= DISPLAYDEVICEINITIALIZED;


/* Thread Execute Phase -- perform I/O and processing */

    DBG( "Entering video_thread_fxn processing loop.\n" );


    Display_get( hDisplay, &procBuf );

    while ( !env->quit )
    {
        Capture_get( hCapture, &cBuf );

        /* Make sure the whole buffer is used for input */
        BufferGfx_resetDimensions(  cBuf);

        if( Venc1_process( encoderHandle, cBuf, encBuf ) != Dmai_EOK )
        {
            ERR( "Encode_video failed in video_thread_fxn\n" );
            status = VIDEO_THREAD_FAILURE;
            break;
        }

        Capture_put( hCapture, cBuf );

        /* Make sure the whole buffer is used for output */
        BufferGfx_resetDimensions( procBuf );
	
        // Encoder may not return data into encBuf if it needs more frames
        if( Buffer_getNumBytesUsed( encBuf ) > 0 )
        {
            ret = Vdec2_process( decoderHandle, encBuf, procBuf );

		    // It's ok to overwrite procBuf because we'll use
		    // Vdec2_getDisplayBuf to get output buffers
            Display_get( hDisplay, &procBuf );
        }		

	    // Positive return values are success codes, negative are failures
	    // Zero is the general success code (EOK)
        if( ret < 0 )
        {
            ERR( "Vdec2_process returned error code: %d\n", ret );
            status = VIDEO_THREAD_FAILURE;
            goto cleanup;
        }

        dBuf = Vdec2_getDisplayBuf( decoderHandle );

        while( dBuf )
        {
            Display_put( hDisplay, dBuf );
            dBuf = Vdec2_getDisplayBuf( decoderHandle );
        }

    }


cleanup:
    /* Thread Delete Phase -- free resources no longer needed by thread */

    DBG("Exited video_thread_fxn processing loop\n");
    DBG("\tStarting video thread cleanup to return resources to system\n");

    /* Use the initMask to only free resources that were allocated     */
    /* Nothing to be done for mixer device, it was closed after init   */

    /* Close video capture device */
    if ( initMask & CAPTUREDEVICEINITIALIZED ) 
    {
        Capture_delete( hCapture );
    }

    /* Close video display device */ 
    if ( initMask & DISPLAYDEVICEINITIALIZED ) 
    {
        Display_delete( hDisplay );
    }

    /* Delete video decoder */
    if ( initMask & VIDEODECODERCREATED )
        Vdec2_delete( decoderHandle );

    /* Delete video encoder */
    if ( initMask & VIDEOENCODERCREATED ) 
        Venc1_delete( encoderHandle );

    /* Close the engine */
    if ( initMask & ENGINEOPENED )
    {
        Engine_close( engineHandle );
    }

    /* Free intermediate (encoded) buffer */
    if ( initMask & ENCODEDBUFFERALLOCATED ) 
    {
        DBG( "\tFreed intermediate video buffer at %p\n", encBuf );
        Buffer_delete( encBuf );
    }

    /* Return the status of the thread execution */
    DBG( "\tVideo thread cleanup complete. Exiting video_thread_fxn.\n" );
    return (void *)status;
}

