# ***********************************************************************************
#
# makefile_profile.mak     Version 0.80        Date: 09-07-08
#
# Revisions:
# - first "standard" GNU makefile for DaVinci workshop
#
# Use:
# - called by parent makefile named "makefile". Currently builds for ARM9 target,
#   however other targets can be supported. User can specify PROFILE (either debug
#   or release or both) when invoking the parent makefile.
#
# - all dependencies (e.g. header files) are handled by the dependency rule.
#
# - uses Configuro to consume packages delivered by TI
#
# - all tools paths are specified in setpaths.mak located two levels above /app.
#
# ***********************************************************************************


# -----------------------
# ------ includes -------
# -----------------------
#
# setpaths.mak includes all absolute paths for DaVinci tools
# and is located two levels above the /app directory.

-include ../../setpaths.mak


# --------------------------------
# ------ user-defined vars -------
# --------------------------------
#
# MVTOOL_PREFIX is specified in setpaths.mak
#
# C_SRCS used to build array of C_OBJS
#

AT := @     # remove the @ symbol to see more feedback from gMake


CC := $(CS_GCC)                              # CC is used to invoke gcc compiler
CC_ROOT := $(CS_INSTALL_PATH)                # CC_ROOT is passed to configuro for building with gcc

DEBUG_CFLAGS := -g -D_DEBUG_
RELEASE_CFLAGS := -O2


CFLAGS := -Wall -fno-strict-aliasing -march=armv5t -D_REENTRANT -I$(CS_INSTALL_PATH)/include -I$(LLIB_INSTALL_DIR)/include -I$(PWD)
LINKER_FLAGS := -lpthread -L$(LLIB_INSTALL_DIR)/lib -lasound 

C_SRCS := $(wildcard *.c)

OBJS := $(subst .c,.o,$(C_SRCS))
C_OBJS = $(addprefix $(PROFILE)/,$(OBJS))

DEPS := $(subst .c,.d,$(C_SRCS))
C_DEPS = $(addprefix $(PROFILE)/,$(DEPS))


# ------------------------------------------
# ------  Configuro related variables ------
# ------------------------------------------

XDCROOT := $(XDC_INSTALL_DIR)
CONFIGURO := $(XDCROOT)/xs xdc.tools.configuro -b $(PWD)/../../config.bld

#export XDCPATH:=/home/user/rtsc_primer/examples;$(XDCROOT)
export XDCPATH:=$(CE_INSTALL_DIR)/packages;$(CE_INSTALL_DIR)/examples;$(XDAIS_INSTALL_DIR)/packages;$(DSPLINK_INSTALL_DIR)/packages;$(CMEM_INSTALL_DIR)/packages;$(BIOS_INSTALL_DIR)/packages;$(BIOSUTILS_INSTALL_DIR)/packages;$(LPM_INSTALL_DIR)/packages;$(FC_INSTALL_DIR)/packages;$(XDCROOT)/packages;$(DMAI_INSTALL_DIR)/packages;$(PWD)/../../myDisplay/packages;$(PWD)/..

PROGNAME := app
CONFIG   := app_cfg

PROFILE  := DEBUG

TARGET   := gnu.targets.arm.GCArmv5T	
PLATFORM := ti.platforms.evm3530	# No AM3517 platform currently

# ------------------------------------------------------------------------
# if you plan to build for these targets, uncomment the appropriate lines
#
# TARGET   := ti.targets.C64
# PLATFORM := ti.platforms.evmEM6446
# TARGET   := gnu.targets.Linux86
# PLATFORM := host.platforms.PC
# ------------------------------------------------------------------------


# -------------------------------------------------
# ----- always keep these intermediate files ------
# -------------------------------------------------
.PRECIOUS : $(C_OBJS)
.PRECIOUS : $(PROFILE)/$(CONFIG)/linker.cmd $(PROFILE)/$(CONFIG)/compiler.opt



# -------------------------------------------------
# --- delete the implicit rules for object files --
# -------------------------------------------------
%.o : %.c


#===========================================================================================================================================

# ----------------------------------------
# -----    build executable (.x)    ------
# ----------------------------------------
$(PROGNAME)_$(PROFILE).xv5T : $(C_OBJS) $(PROFILE)/$(CONFIG)/linker.cmd 
	@echo; echo "1.  ----- Need to generate executable file: $@ "
	@$(CC) $(CFLAGS) $(LINKER_FLAGS) $^ -o $@
	@echo "          Successfully created executable : $@ " ; echo


# ----------------------------------------
# -----    Object File Rule (.o)    ------
# ----------------------------------------
$(PROFILE)/%.o : %.c $(PROFILE)/$(CONFIG)/compiler.opt 
	@echo
	@echo "2.  ----- Need to generate:      $@ (due to: $(wordlist 1,1,$?) ...)"
	@mkdir -p $(dir $@)
	$(AT) $(CC) $(CFLAGS) $($(PROFILE)_CFLAGS) $(shell cat $(PROFILE)/$(CONFIG)/compiler.opt) -c $< -o $@
	@echo "          Successfully created:  $@ "


# ---------------------------------------
# -----    Dependency Rule (.d)    ------
# ---------------------------------------
$(PROFILE)/%.d : %.c $(PROFILE)/$(CONFIG)/compiler.opt
	@echo "3.  ----- Need to generate dep info for:        $< "
	@echo "          Generating dependency file   :        $@ "
	@mkdir -p $(PROFILE)
	@$(AT) $(CC) -MM $(CFLAGS) $($(PROFILE)_CFLAGS) $(shell cat $(PROFILE)/$(CONFIG)/compiler.opt) $<   > $@

	@echo "          Formatting dependency file:           $@ "
	@$(AT) $(call format_d ,$@,$(PROFILE)/)
	@echo "          Dependency file successfully created: $@ " ; echo


# ----------------------------------------
# -----    Configuro Rule (.cfg)    ------
# ----------------------------------------
$(PROFILE)/%/linker.cmd $(PROFILE)/%/compiler.opt : %.cfg 
	@echo "4.  ----- Starting Configuro for $^  (note, this may take a minute)"
     ifdef DUMP
	$(AT) $(CONFIGURO) -c $(CC_ROOT) -t $(TARGET) -p $(PLATFORM) -r $(PROFILE) -o $(PROFILE)/$(CONFIG) $<
     else
	@mkdir -p $(PROFILE)/$(CONFIG)
	$(AT) $(CONFIGURO) -c $(CC_ROOT) -t $(TARGET) -p $(PLATFORM) -r $(PROFILE) -o $(PROFILE)/$(CONFIG) $< > $(PROFILE)/$(CONFIG)_results.log
     endif
	@echo "          Configuro has completed; it's results are in $(CONFIG) " ; echo


# This additional rule creates an empty config file if one doesn't already exist
%.cfg :
	touch $(CONFIG).cfg


# ----------------------------
# -----    build all    ------
# ----------------------------
#
# By default, "all" builds both debug and release versions
#

.PHONY  : all 
all : $(PROGNAME)_$(PROFILE).xv5T
	@echo ; echo "The target ($<) has been built." 
	@echo


# ----------------------------
# -----    clean all    ------
# ----------------------------
#
# "clean" removes the intermediate directories, the local copies
# of the executables and the installed executables (using EXEC_DIR).
# EXEC_DIR is specified in setpaths.mak two levels above /app.
# "clean" removes all files created by BOTH profiles (DEBUG & RELEASE)
# regardless of which profile was selected.
#

.PHONY : clean
clean  : 
	@echo ; echo "--------- Cleaning up files for $(PROFILE) -----"
	rm -rf $(PROFILE)
	rm -rf $(PROGNAME)_$(PROFILE).xv5T
	rm -rf $(EXEC_DIR)/$(PROGNAME)_$(PROFILE).xv5T 
	rm -rf $(C_DEPS)
	rm -rf $(C_OBJS)
	@echo



# --------------------------
# -----    install    ------
# --------------------------
#
# copies the executables to the shared filesys for the DaVinci EVM
# EXEC_DIR is specified in setpaths.mak two levels above /app.
#

.PHONY  : install 
install : $(PROGNAME)_$(PROFILE).xv5T $(CE_INSTALL_DIR)/examples/ti/sdo/ce/examples/servers/all_codecs/bin/ti_platforms_evm3530/all.x64P ../osdfiles/ti_rgb24_480x60.bmp
	@echo 
	@echo  "0.  ----- Installing $(PROGNAME)_$(PROFILE).xv5T to 'Execution Directory' -----"
	@echo  "          Execution Directory:  $(EXEC_DIR)"
	@mkdir -p $(EXEC_DIR)
	@cp    $^ $(EXEC_DIR)
	@echo  "          Install (i.e. copy) has completed" ; echo


# --------------------------
# -----    server     ------
# --------------------------
#
# builds the DSP server
#

# Not the proper use of .PHONY, but this will force make to
#      try to rebuild the server each time
# The runserver.sh script, however, will correctly gauge
#      which steps in partial build must be rerun

.PHONY: ../server/server_debug.x64P ../server/server_release.x64P

../server/server_debug.x64P ../server/server_release.x64P: 
	@echo "            Rebuilding DSP server"
	@cd ..;./runserver.sh

# ------------------------------------------------------------
# ------------------    Macro: format_d     ------------------
# ------------------------------------------------------------
define format_d
 @# echo " Formatting dependency file: $@ "
 @# echo " This macro has two parameters: "
 @# echo "   Dependency File (.d): $1     "
 @# echo "   Profile: $2                  "
 @mv -f $1 $1.tmp
 @echo '$$(warning --- Reading from included file: $1 ---)' > $1
 @sed -e 's|.*:|$2$*.o:|' < $1.tmp >> $1
 @rm -f $1.tmp
endef



# ---------------------------------------------------------
# -----    Including goal specific files     --------------
# ---------------------------------------------------------
ifneq ($(filter clean,$(MAKECMDGOALS)),clean)
  -include $(C_DEPS)
endif



# -------------------------------------------------------
# -----    Print out build information    ---------------
# -------------------------------------------------------
ifdef DUMP
  $(warning To view build commands, invoke make with argument 'AT= ')
  $(warning To view build variables, invoke make with 'DUMP=1')

  $(warning Source Files: $(C_SRCS))
  $(warning Object Files: $(C_OBJS))
  $(warning Depend Files: $(C_DEPS))

  $(warning Base program name : $(PROGNAME))
  $(warning Configuration file: $(CONFIG))
  $(warning Make Goals        : $(MAKECMDGOALS))

  $(warning Xdcpath :  $(XDCPATH))
  $(warning Target  :  $(TARGET))
  $(warning Platform:  $(PLATFORM))
endif




